import numpy as np
import matplotlib.pyplot as plt
import copy
import jax.numpy as jnp
from jax import jit
import pennylane as qml
from ..toolboxes.toolbox_statevector import distance, fidelity


class Circuit_P:
    """
    The Pennylane version of the circuit stack consists of 3 levels, which from bottom to top are: gate, block and
    circuit.

    Circuits are at the top level of the stack.

    Some important attributes of a circuit include:

    1. nq: The number of qubits involved in the circuit.
    2. components: A list of circuits of the constituent layer(s).
    3. param_shape: The shape of parameters for each component.
    4. mq: "measure_qubits". By default, it is False, meaning all qubits are measured and the exact output statevector
       will be returned. Else, it could be a list of the indices of the qubits to be measured. Then the probability
       distribution of bitstrings of the output qubits will be returned.

    :param nq: the number of qubits
    :type nq: int
    :param mq: the qubits to be measured
    :type mq: list of integers or False
    """
    def __init__(self, nq=4, **kwargs):
        self.nq = nq
        self.components = []
        self.param_shape = []
        self.mq = kwargs.get('mq', False)  # False means measure all qubits
        self.circuit, self.circuit_pre = None, None
        self.device = qml.device("default.qubit.jax", wires=self.nq)

    def __repr__(self):
        return "Circuit_P({})".format(self.nq)

    def __str__(self):
        return "PC-{}Q_".format(self.nq)

    def __add__(self, other):
        """
        The components of the two circuits will be added together, and the "mq" of the circuit will be given by the
        other circuit.

        :param other: circuit to be added
        :type other: Circuit_P
        :return: new circuit
        :rtype: Circuit_P
        """
        new_circuit = copy.deepcopy(other)
        new_circuit.param_shape = self.param_shape + other.param_shape
        new_circuit.components = self.components + other.components
        new_circuit.assemble()
        return new_circuit

    def get_np(self):
        """
        Compute the total number of parameters in the block.

        :rtype: int
        """
        return sum([np.prod(shape) for shape in self.param_shape])

    def params_to_proper_shape(self, flattened_params):
        """
        Reshape a flattened array of parameters into the form ready to be fed into the circuit.

        :param flattened_params: flattened parameters
        :type flattened_params: (jax) numpy array / list of float
        :return: parameters reshaped according to param_shape
        :rtype: list of (jax) numpy arrays
        """
        reshaped_params = []
        index = 0
        for shape in self.param_shape:
            length = np.prod(shape)
            reshaped_params.append(flattened_params[index: index + length].reshape(shape))
            index += length
        return reshaped_params

########################################################################################################################
    def assemble(self):
        """
        Automatically called during initialization to generate the circuit function.
        """
        self.__assemble_1()
        self.__assemble_2()

    def __assemble_1(self):
        """
        Automatically called during initialization to generate the 'circuit-pre' function, which outputs the statevector
        in linear array form. 'circuit-pre' is called, for example, in 'plot'.
        """
        @qml.qnode(self.device, diff_method="backprop", interface="jax")
        def circuit_pre(params):
            for i in range(len(self.components)):
                self.components[i](params[i])
            if self.mq:
                return qml.probs(wires=self.mq)
            return qml.state()

        self.circuit_pre = circuit_pre

    def __assemble_2(self):
        """
        Automatically called during initialization to generate the 'circuit' function, which executes the "circuit_pre"
        function but additionally reshape the output statevector into tensor form. In the case mq is False, as the
        output of "circuit-pre" is a bitstring distribution instead of array, we leave "circuit" to be the same as
        "circuit-pre".
        """
        @jit
        def circuit_psi(params):
            nq_out = len(self.mq) if self.mq else self.nq
            return self.circuit_pre(params).reshape(nq_out * [2])

        if self.mq:
            self.circuit = self.circuit_pre
        else:
            self.circuit = circuit_psi

########################################################################################################################
    def run_with_param_input(self, params):
        """
        Run the full circuit with parameter inputs for the parametrized gates.

        :param params: parameter inputs
        :type params: (jax) numpy array / list
        """
        return self.circuit(params)

    def plot(self, file_name=None, **kwargs):
        """
        Make a plot of the circuit and optionally save as a file.

        :param file_name: The name of the file that saves the plot
        :type file_name: str
        :param fig_size: size of the image
        :type fig_size: tuple of int
        :param dpi: quality of the image
        :type dpi: int
        """
        qml.drawer.use_style("black_white")
        params = kwargs.get("params", [np.zeros(param_shape) for param_shape in self.param_shape])
        fig, ax = qml.draw_mpl(self.circuit_pre, expansion_strategy="device")(params)
        fig.set_size_inches(kwargs.get("fig_size", (20, 10)))
        if file_name is not None:
            fig.savefig(file_name, dpi=kwargs.get("dpi", 100))
            plt.close()
        else:
            plt.show()

    def to_qasm(self, params, file_name):
        """
        Outputs the gate sequence with parameter inputs in OpenQASM format. May fail, depending on the type of gates
        involved.

        :param params: parameter inputs
        :type params: (jax) numpy array / list
        :param file_name: The name of the file that saves the plot
        :type file_name: str
        """
        self.circuit_pre(params)
        text = self.circuit_pre.qtape.to_openqasm()
        file = open(file_name, "w+")
        file.write(text)
        file.close()

########################################################################################################################
    def get_loss(self, params, psi, lf='td'):
        """
        Applicable when mq is False.

        Compute the "loss" of the output state of the VQC with parameter inputs compared to a target state psi.

        Often called during optimization for state preparation. Could also be used for more general purposes.

        :param params: parameter inputs
        :type params: (jax) numpy array / list
        :param psi: target state
        :type psi: (jax) numpy array
        :param lf: choice of loss function, currently supports:

                      1. coordinate-wise distance
                      2. squared coordinate-wise distance
                      3. 1 - fidelity
                      4. trace distance
        :type lf: int
        """
        if lf == 'd':  # Vector distance
            return self.get_distance_with(params, psi)
        if lf == 'd2':  # Vector distance squared
            return self.get_distance_with(params, psi)**2
        if lf == "1-f":  # 1 - fidelity
            return 1 - self.get_fidelity_with(params, psi)
        if lf == "td":  # Trace distance, related to fidelity for pure states
            return 2 * jnp.sqrt(1 - self.get_fidelity_with(params, psi))

    def get_distance_with(self, params, psi):
        """
        Compute the coordinate-wise distance between the output state of the VQC with parameter inputs and a target
        state psi.

        :param params: parameter inputs
        :type params: (jax) numpy array / list
        :param psi: target state
        :type psi: (jax) numpy array
        """
        psi_out = self.run_with_param_input(params)
        return distance(psi_out, psi)

    def get_fidelity_with(self, params, psi):
        """
        Compute the fidelity between the output state of the VQC with parameter inputs and a target state psi.

        :param params: parameter inputs
        :type params: (jax) numpy array / list
        :param psi: target state
        :type psi: (jax) numpy array
        """
        psi_out = self.run_with_param_input(params)
        return fidelity(psi_out, psi)

########################################################################################################################


import numpy as np
import matplotlib.pyplot as plt
import jax


class Expressibility_Evaluation:
    """
    Expressibility is a popular descriptor of VQCs. It compares the output statevectors of a VQC with Haar random states
    of the corresponding Hilbert space. It is computed as the KL-divergence between two histograms. For more details,
    please refer to:

    Reference: Expressibility and entangling capability of parameterized quantum circuits for hybrid quantum-classical
    algorithms. S. Sim et al. arXiv: 1905.10876

    :param circuit: the VQC to be evaluated
    :type circuit: Circuit object from circuit_stack or Circuit_P object from circuit_stack_pennylane (recommended)
    :param n_samples: the number of pairs of statevectors to sample, 10000 by default
    :type n_samples: int
    :param n_bins: the number of bins for the histogram, 75 by default
    :type n_bins: int
    :param seed: the random seed for the initialization of parameters, 1 by default
    :type seed: jax random seed
    :param parameter_bounds: the interval in which parameters are sampled, [0, 1] by default
    :type parameter_bounds: [float, float]
    """
    def __init__(self, circuit, **kwargs):
        self.circuit = circuit
        self.n_samples = kwargs.get('n_samples', 10000)
        self.n_bins = kwargs.get('n_bins', 75)
        self.seed = kwargs.get("seed", 1)
        self.parameter_bounds = kwargs.get("parameter_bounds", [0, 1])
        self.parameter_samples = None
        self.fidelity_samples = None
        self.bin_edges = None
        self.bin_heights = None
        self.bin_heights_harr = None
        self.DKL = None

    def __repr__(self):
        return "Expressibility_Evaluation({})".format(self.circuit)

    def __str__(self):
        return "Expressibility_Evaluation" + str(self.circuit)

    def __get_parameter_samples(self):
        np = self.circuit.get_np()
        key = jax.random.PRNGKey(self.seed)
        parameter_samples = jax.random.uniform(key, shape=tuple([np * 2 * self.n_samples]),
                                               minval=self.parameter_bounds[0], maxval=self.parameter_bounds[1])
        # First get the parameters in an array, and then divide them into 2 * n_sample portions.
        parameter_samples = [parameter_samples[i: i + np] for i in range(0, len(parameter_samples), np)]
        if hasattr(self.circuit, "param_shape"):  # For Pennylane circuits, we need to reshape the parameters
            parameter_samples = [self.circuit.params_to_proper_shape(params) for params in parameter_samples]
        self.parameter_samples = parameter_samples

    def __get_fidelity_samples(self):
        fidelity_samples = []
        for i in range(self.n_samples):
            psi0 = self.circuit.run_with_param_input(self.parameter_samples[2 * i])  # statevector of the first sample
            fidelity = self.circuit.get_fidelity_with(self.parameter_samples[2 * i + 1], psi0)
            # compare with statevector of the second sample
            fidelity_samples.append(np.array(fidelity))
        self.fidelity_samples = np.array(fidelity_samples)
        self.fidelity_samples.reshape(-1)  # flatten to an array

    def __get_histogram_counts(self):
        bin_edges = np.linspace(0, 1, self.n_bins + 1)
        weights = np.ones_like(self.fidelity_samples) / len(self.fidelity_samples)
        self.bin_heights, self.bin_edges = np.histogram(self.fidelity_samples, bins=bin_edges, weights=weights)

    def __compute_Harr(self):
        N = 2 ** self.circuit.nq
        n_bins = len(self.bin_edges) - 1
        self.bin_heights_harr = np.clip(np.array([(1 - self.bin_edges[i]) ** (N - 1) -
                                                  (1 - self.bin_edges[i + 1]) ** (N - 1)
                                                  for i in range(n_bins)]), 1e-300, 1)

    def __compute_DKL(self):
        self.DKL = np.sum(self.bin_heights * np.log(np.clip(self.bin_heights / self.bin_heights_harr, 1e-300, 1e300)))

    def __plot(self):
        fig, ax = plt.subplots(1, 1)
        bin_centers = (self.bin_edges + self.bin_edges[1] / 2)[:-1]
        width = 2 * bin_centers[0]
        ax.bar(bin_centers, self.bin_heights, width=width, label=str(self.circuit), color="#226E9C", linewidth=0)
        ax.plot(bin_centers, self.bin_heights_harr, label="Harr", color="#0D4A70")
        ax.legend(loc="upper right")
        ax.text(0.4, 0.95, "DKL = " + str(round(self.DKL, 3)), ha="right", va="center", transform=ax.transAxes)
        ax.set_xlabel("Fidelity")
        ax.set_ylabel("Frequency")
        ax.set_title("Fidelity between states randomly sampled from parameter space, " + str(self.n_samples) + " pairs")
        plt.show()

    def run(self):
        """
        Call this function to start the task.
        """
        self.__get_parameter_samples()
        self.__get_fidelity_samples()
        self.__get_histogram_counts()
        self.__compute_Harr()
        self.__compute_DKL()
        self.__plot()





class Block:
    """
    The circuit stack consists of 4 levels, which from bottom to top are: gate, block, layer and circuit.

    A block is the repeating unit in a layer. A block consists of >= 1 parameterized or fixed gate(s).
    The most important attributes of a block are three lists, called "gs", "il" and "npl".

    1. "gs" stands for gate sequence,
        which is a list of strings (names of the gates). The functions realizing the gates
        are defined in the file gate.py. The strings will be mapped to the functions at the circuit level.
    2. "il" stands for index list,
        which is a list of lists of integers called 'local indices', which defines the qubits on which
        the gates apply.
    3. "npl" stands for number of parameter list,
        which is a list storing the number of parameters each gate takes.
    """
    def __init__(self):
        self.gs, self.il, self.npl = [], [], []
        # npl stands for "number-of-parameters list", which stores the number of parameters each gate in the block takes
        # gs stands for "gate_sequence", which stores the name of each gate in the block, which will eventually be used
        # to call the gate functions at the circuit level.
        # il stands for "index_list",
        self.assemble()

    def __repr__(self):
        return "Block()"

    def __str__(self):
        return "B_"

    def assemble(self):
        """
        Automatically called during initialization to generate the three attributes.
        """
        pass

    def get_np(self):
        """
        Compute the total number of parameters in the block.
        """
        return sum(self.npl)


class Block_GU1(Block):
    """
    This block consists of only 1 general single qubit unitary. It's takes 3 parameters.
    """
    def __init__(self):
        super(Block_GU1, self).__init__()

    def __repr__(self):
        return "Block_GU1()"

    def __str__(self):
        return "B-GU1_"

    def assemble(self):
        self.gs = ['GU1']
        self.il = [[0]]
        self.npl = [3]


class Block_GU2(Block):
    """
    This block consists of only 1 general two qubit unitary (GU2). It takes 15 parameters.
    """
    def __init__(self):
        super(Block_GU2, self).__init__()

    def __repr__(self):
        return "Block_GU2()"

    def __str__(self):
        return "B-GU2_"

    def assemble(self):
        self.gs = ['GU2']
        self.il = [[0, 1]]
        self.npl = [15]


class Block_CZ(Block):
    """
    This block consists of a Controlled-Z gate, followed by 2 general single qubit unitaries, one on each qubit. It
    takes 6 parameters.
    """
    def __init__(self):
        super(Block_CZ, self).__init__()

    def __repr__(self):
        return "Block_CZ()"

    def __str__(self):
        return "B-CZ_"

    def assemble(self):
        self.gs = ['CZ', 'GU1', 'GU1']
        self.il = [[0, 1], [0], [1]]
        self.npl = [0, 3, 3]


class Block_CNOT(Block):
    """
    This block consists of a Controlled-X gate, followed by 2 general single qubit unitaries (GU1), one on each qubit.
    It takes 6 parameters. As an example, the three attributes for this block writes:

    1. gs = ['CNOT', 'GU1', 'GU1']
    2. il = [[0, 1], [0], [1]]
    3. npl = [0, 3, 3]
    """
    def __init__(self):
        super(Block_CNOT, self).__init__()

    def __repr__(self):
        return "Block_CNOT()"

    def __str__(self):
        return "B-CNOT_"

    def assemble(self):
        self.gs = ['CNOT', 'GU1', 'GU1']
        self.il = [[0, 1], [0], [1]]
        self.npl = [0, 3, 3]


class Layer:
    """
    The circuit stack consists of 4 levels, which from bottom to top are: gate, block, layer and circuit.

    A layer is a component of the circuit, which could repeat itself a few times in the circuit. A layer
    is an arrangement of repeating blocks that covers all qubits.

    The most important attributes of a circuit are also "gs", "il" and "npl", which are formed by concatenating the ones
    of its block components.

    :param block: The repeating block in the layer
    :type block: Block
    :param nq: The number of qubits
    :type nq: int
    """
    def __init__(self, block, **kwargs):
        self.block = block
        self.nq = kwargs.get('nq', 4)  # The number of qubits.
        self.gs, self.il, self.npl = [], [], []
        self.assemble()

    def __repr__(self):
        return "Layer({})".format(self.block)

    def __str__(self):
        pass

    def assemble(self):
        """
        Automatically called during initialization to generate the three attributes.
        """
        pass

    def get_np(self):
        """
        Compute the total number of parameters in the block.
        """
        return sum(self.npl)


class RCMPSLayer(Layer):
    """
    A ladder-shaped layer that grows in one direction, inspired by MPS of right canonical form.
    """
    def __init__(self, block, **kwargs):
        super(RCMPSLayer, self).__init__(block, **kwargs)

    def __repr__(self):
        return "RCMPSLayer({})".format(self.block)

    def assemble(self):
        for global_index in range(self.nq-1):
            self.gs += self.block.gs
            for local_index in self.block.il:
                self.il.append([i + global_index for i in local_index])
            self.npl += self.block.npl


class PRCMPSLayer(Layer):
    """
    A ladder-shaped layer that grows in one direction, inspired by MPS of right canonical form with periodic boundary
    condition.
    """
    def __init__(self, block, **kwargs):
        super(PRCMPSLayer, self).__init__(block, **kwargs)

    def __repr__(self):
        return "PRCMPSLayer({})".format(self.block)

    def assemble(self):
        for global_index in range(self.nq):
            self.gs += self.block.gs
            for local_index in self.block.il:
                self.il.append([(i + global_index) % self.nq for i in local_index])
            self.npl += self.block.npl


class MCMPSLayer(Layer):
    """
    A ladder-shaped layer that grows in both directions, inspired by MPS of mixed canonical form.
    """
    def __init__(self, block, **kwargs):
        super(MCMPSLayer, self).__init__(block, **kwargs)

    def __repr__(self):
        return "MCMPSLayer({})".format(self.block)

    def assemble(self):
        pass


class IRLayer(Layer):
    """
    A flat layer of single qubit unitaries. Typically used for initial rotations on the qubits.
    """
    def __init__(self, block, **kwargs):
        super(IRLayer, self).__init__(block, **kwargs)

    def __repr__(self):
        return "IRLayer({})".format(self.block)

    def assemble(self):
        for global_index in range(self.nq):
            self.npl += self.block.npl
            self.gs += self.block.gs
            self.il.append(([global_index]))



from tqdm import tqdm
import time
import numpy as np
import jax
import jax.numpy as jnp
import optax
from ..circuit_stack.gate import Haar_Random
from ..toolboxes.toolbox_statevector import fidelity


class Compression_Sweeping:
    """
    The goal of compression is to iteratively optimize the parameters of the VQC, such that it outputs a statevector
    with high fidelity compared to the target state.

    This task uses a sweeping optimization method, inspired by some tensor network techniques.
    Ref. Real- and Imaginary-Time Evolution with Compressed Quantum Circuits. PRX Quantum 2, 010342

    :param psi_target: The target state to prepare
    :type psi_target: (jax) numpy array in tensor shape
    :param circuit: the VQC to be trained for state preparation
    :type circuit: Circuit object from circuit_stack. Usage with circuit_stack_Pennylane is not supported
    :param steps: the maximal number of optimization iterations, 1e3 by default
    :type steps: int
    :param ea: the maximal absolute error before stopping the optimization, 1e-10 by default
    :type ea: float
    :param er: the maximal relative error before stopping the optimization, 1e-10 by default
    :type er: float
    :param seed: the random seed for the initialization of parameters, 1 by default
    :type seed: jax random seed
    :param init_filled_circuit: initial values for the parameterized gates, None by default and random initialization
           will be triggered.
    :type init_filled_circuit: list of (jax) numpy arrays or None
    """

    def __init__(self, psi_target, circuit, **kwargs):
        self.psi_target = psi_target
        self.circuit = circuit

        self.optimizer = "Sweeping"
        self.steps = kwargs.get("steps", 1e3)  # The maximal iteration number,
        # to be overwritten later by the actual iteration number
        self.ea = kwargs.get("ea", 1e-10)  # The absolute convergence error
        self.er = kwargs.get("er", 1e-10)  # The relative convergence error
        self.seed = kwargs.get("seed", 1)  # Random seed

        self.circuit.filled_circuit = kwargs.get("init_filled_circuit", None)
        if self.circuit.filled_circuit is None:
            self.__initialize_filled_circuit()

        self.record = None

    def __repr__(self):
        return "Compression_Sweeping({},{})".format(self.psi_target, self.circuit)

    def __str__(self):
        return "Compression_Sweeping_" + str(self.circuit)

    def __initialize_filled_circuit(self):
        """
        Initialize a filled_circuit with Haar random unitaries.
        """
        key = jax.random.PRNGKey(self.seed)
        self.circuit.fill_params(np.ones(self.circuit.get_np()))  # just to create a trivial filled_circuit first
        for ig in self.circuit.get_where_parametrized():
            global_index = self.circuit.il[ig]
            nq = len(global_index)
            key, subkey = jax.random.split(key)
            U = Haar_Random(nq, subkey).reshape(*tuple(2 * len(global_index) * [2]))
            self.circuit.filled_circuit[ig] = U

    def run(self):
        """
        Call this function to start the task.
        """
        self.record = {'optimized_gates': [], 'fidelity': []}

        psi_original = self.psi_target
        psi_in = np.zeros(self.circuit.nq * [2])
        psi_in[tuple(self.circuit.nq * [0])] = 1

        step = 0
        absolute_error = 1
        relative_error = 1

        while step < self.steps and absolute_error > self.ea and relative_error > self.er:
            step += 1

            for ig in self.circuit.get_where_parametrized():

                global_index = self.circuit.il[ig]

                unitary_shape = tuple(2 * [2 ** len(global_index)])
                tensor_shape = tuple(2 * len(global_index) * [2])

                E = self.circuit.environment_contraction(psi_in, psi_original.conjugate(),
                                                         ig).reshape(*unitary_shape)
                X, Sigma, Ydagger = jnp.linalg.svd(E)
                U = jnp.transpose(jnp.conjugate(X @ Ydagger))

                self.circuit.filled_circuit[ig] = U.reshape(*tensor_shape)

            psi_out = self.circuit.forward_contraction(psi_in)
            f = fidelity(psi_out, psi_original)
            self.record['fidelity'].append(f)

            pre_error = absolute_error
            absolute_error = 1 - f  # Error is defined by 1 - fidelity
            relative_error = np.abs(absolute_error - pre_error) / np.abs(pre_error)

        self.record['optimized_gates'] = [self.circuit.filled_circuit[ig]
                                          for ig in self.circuit.get_where_parametrized()]
        # only store parametrized gates

        self.steps = step  # replace the maximum step by the actual used step


class Compression_Adam:
    """
    The goal of compression is to iteratively optimize the parameters of the VQC, such that it outputs a statevector
    with high fidelity compared to the target state.

    This task uses Adam optimizer, a popular gradient based optimization method for machine learning and optimization.

    :param psi_target: The target state to prepare
    :type psi_target: (jax) numpy array in tensor shape
    :param circuit: the VQC to be trained for state preparation
    :type circuit: Circuit object from circuit_stack or Circuit_P object from circuit_stack_pennylane (recommended)
    :param steps: the maximal number of optimization iterations, 1e3 by default
    :type steps: int
    :param lr: the learning rate
    :type lr: float
    :param lf: the loss function. See circuit.py from circuit_stack or circuit_stack_pennylane
    :type lf: float
    :param seed: the random seed for the initialization of parameters, 1 by default
    :type seed: jax random seed
    :param init_params: initial values for the parameterized gates, None by default and random initialization
           will be triggered.
    :type init_params: (jax) numpy array
    """

    def __init__(self, psi_target, circuit, **kwargs):
        self.psi_target = psi_target
        self.circuit = circuit

        self.optimizer = "Adam"
        self.lr = kwargs.get("lr", 0.03)  # The learning rate
        self.optimizer_function = optax.adam(self.lr)  # By default, use Adam optimizer
        self.steps = kwargs.get('steps', 5000)  # The number of iterations
        self.seed = kwargs.get("seed", 1)  # Random seed
        self.lf = kwargs.get("lf", "1-f")  # Loss function

        self.init_params = kwargs.get("init_params", None)
        if self.init_params is None:
            self.initialize_params()

        self.record = None

    def __repr__(self):
        return "Compression_Adam({},{})".format(self.psi_target, self.circuit)

    def __str__(self):
        return "Compression_Adam_" + str(self.circuit)

    def initialize_params(self):
        """
        Initialize a filled_circuit with Haar random unitaries.
        """
        key = jax.random.PRNGKey(self.seed)
        self.init_params = jax.random.uniform(key, shape=tuple([self.circuit.get_np()]))
        if hasattr(self.circuit, "param_shape"):  # For Pennylane circuits, we need to reshape the parameters
            self.init_params = self.circuit.params_to_proper_shape(self.init_params)

    def run(self):
        """
        Call this function to start the task.
        """
        self.record = {'params': [], 'loss': [], 'fidelity': [], 'time': 0}

        def optimization_step(params, state, psi_original):
            """
            The iterative step in the optimization
            :param params: The parameters of the gates
            :param state: current state
            :param psi_original: the reference state, FRQI state
            :return:
            """
            grads = jax.grad(self.circuit.get_loss)(params, psi_original, lf=self.lf)
            # computes the gradient of the loss function at the current params
            updates, state = self.optimizer_function.update(grads, state)
            # opt.update outputs the computed gradient updates and the new state
            params = optax.apply_updates(params, updates)
            return params, state

        params = self.init_params
        state = self.optimizer_function.init(self.init_params)
        psi_original = self.psi_target

        start_time = time.monotonic()
        for _ in tqdm(range(self.steps)):
            params, state = optimization_step(params, state, psi_original)
            self.record['loss'].append(self.circuit.get_distance_with(params, psi_original))  # compute loss (distance)
            self.record['fidelity'].append(self.circuit.get_fidelity_with(params, psi_original))  # compute overlap
        end_time = time.monotonic()

        self.record['time'] = round((end_time - start_time) / self.steps, 3)

        self.record['params'] = np.concatenate([np.ravel(component) for component in params])  # flatten



from .circuit import Circuit
from .layer import *
from .block import *


class MPS(Circuit):
    """
    This is the basis class for all circuits inspired by matrix product states (MPS).

    :param layer: The layer object underlying the circuit.
    :type layer: Layer
    :param nl: The number of layers in the circuit.
    :type nl: int
    :param nq: The number of qubits in the circuit.
    :type nq: int
    :param irb: The block object for initial rotation, if applicable. It's None by default.
    :type irb: Block
    """
    def __init__(self, layer, nl=1, nq=4, irb=None, **kwargs):
        super(MPS, self).__init__(nq, **kwargs)
        self.layer = layer
        self.nb = self.nq - 1
        self.nl = nl
        self.irb = irb
        self.components = [self.layer] * nl

        if self.irb is not None:
            self.components = [IRLayer(self.irb)] + self.components

        self.assemble()

    def __repr__(self):
        return "MPS({}, {}, {}, {})".format(repr(self.layer), self.nl, self.nq, repr(self.irb))

    def __str__(self):
        return "C-{}L{}Q_".format(self.nl, self.nq) + str(self.layer.block)


class MPS_GU2(MPS):
    """
    MPS with GU2 blocks.

    :param version: define the block arrangement in:

                    0. right canonical form
                    1. right canonical form with periodic condition
                    2. mixed canonical form

    :type version: int

    """
    def __init__(self, nl=1, nq=4, version=0, **kwargs):
        if version == 1:
            layer = MCMPSLayer(Block_GU2(), nq=nq)
        elif version == 2:
            layer = PRCMPSLayer(Block_GU2(), nq=nq)
        else:
            layer = RCMPSLayer(Block_GU2(), nq=nq)
        self.version = version
        super(MPS_GU2, self).__init__(layer, nl, nq, **kwargs)

    def __repr__(self):
        return "MPS_GU2({}, {}, {})".format(self.nl, self.nq, self.version)


class MPS_CZ(MPS):
    """
    MPS with CZ blocks.

    :param pm1: parameterization method for GU1. See gate.py
    :type pm1: int
    :param version: define the block arrangement in:

                    0. right canonical form
                    1. right canonical form with periodic condition
                    2. mixed canonical form

    :type version: int

    """
    def __init__(self, nl=1, nq=4, irb=None, pm1=0, version=0, **kwargs):
        if version == 1:
            layer = MCMPSLayer(Block_CZ(), nq=nq)
        elif version == 2:
            layer = PRCMPSLayer(Block_CZ(), nq=nq)
        else:
            layer = RCMPSLayer(Block_CZ(), nq=nq)
        self.version = version
        super(MPS_CZ, self).__init__(layer, nl, nq, irb, **kwargs)
        self.pmd['GU1'] = pm1

    def __repr__(self):
        return "MPS_CZ({}, {}, {}, {}, {})".format(self.nl, self.nq, repr(self.irb), self.pmd['GU1'], self.version)


class MPS_CNOT(MPS):
    """
    MPS with CNOT blocks.

    :param pm1: parameterization method for GU1. See gate.py
    :type pm1: int
    :param version: define the block arrangement in:

                    0. right canonical form
                    1. right canonical form with periodic condition
                    2. mixed canonical form

    :type version: int
    """
    def __init__(self, nl=1, nq=4, irb=None, pm1=0, version=0, **kwargs):
        if version == 1:
            layer = MCMPSLayer(Block_CNOT(), nq=nq)
        elif version == 2:
            layer = PRCMPSLayer(Block_CNOT(), nq=nq)
        else:
            layer = RCMPSLayer(Block_CNOT(), nq=nq)
        self.version = version
        super(MPS_CNOT, self).__init__(layer, nl, nq, irb, **kwargs)
        self.pmd['GU1'] = pm1

    def __repr__(self):
        return "MPS_CNOT({}, {}, {}, {}, {})".format(self.nl, self.nq, repr(self.irb), self.pmd['GU1'], self.version)



import numpy as np
from .circuit import Circuit_P
from .block import *


class MPS_P(Circuit_P):
    """
    This is the basis class for all circuits inspired by matrix product states (MPS).

    :param block: The block object underlying the circuit.
    :type block: Block_P
    :param nl: The number of layers in the circuit.
    :type nl: int
    :param nq: The number of qubits in the circuit.
    :type nq: int
    :param irb: The block object for initial rotation, if applicable. It's None by default.
    :type irb: Block
    :param version: define the block arrangement in:

                 0. right canonical form
                 1. right canonical form with periodic condition
                 2. mixed canonical form

    :type version: int
    """
    def __init__(self, block, nl=1, nq=11, irb=None, version=0, **kwargs):
        super(MPS_P, self).__init__(nq, **kwargs)
        self.block = block
        self.nl = nl
        self.nb = self.nq - 1
        self.irb = irb
        self.version = version

        def initial_rotation(params):
            for q in range(self.nq):
                self.irb.circuit(params[q], q)

        if version == 1:  # right canonical form with periodic boundary condition
            def ladder_layers(params):
                for layer in range(self.nl):
                    for i in range(self.nq):
                        p = params[layer][i]
                        self.block.circuit(p, (i, (i+1) % self.nq))

            self.param_shape = [(self.nl, self.nq, self.block.np)]

        elif version == 2:  # mixed canonical form
            def ladder_layers(params):
                for layer in range(self.nl):
                    mi = int(np.ceil(self.nq / 2)) - 1  # middle index
                    for i in range(mi):
                        pr = params[layer][2*i]
                        self.block.circuit(pr, (mi + i, mi + i + 1))
                        pl = params[layer][2*i + 1]
                        self.block.circuit(pl, (mi - 1 - i, mi - i))
                    if self.nq % 2 == 0:
                        pr = params[layer][-1]
                        self.block.circuit(pr, (self.nq - 2, self.nq - 1))

            self.param_shape = [(self.nl, self.nb, self.block.np)]

        else:
            def ladder_layers(params):
                for layer in range(self.nl):
                    qml.MPS(
                        wires=range(self.nq),
                        n_block_wires=2,
                        block=self.block.circuit,
                        n_params_block=self.block.np,
                        template_weights=params[layer],
                    )

            self.param_shape = [(self.nl, self.nb, self.block.np)]

        if self.irb is not None:
            self.param_shape = [(self.nq, 3)] + self.param_shape
            self.components = [initial_rotation, ladder_layers]
        else:
            self.components = [ladder_layers]

        self.assemble()

    def __repr__(self):
        return "MPS_P({}, {}, {}, {}, {})".format(repr(self.block), self.nl, self.nq, repr(self.irb), self.version)

    def __str__(self):
        return "PC-{}L{}Q_".format(self.nl, self.nq) + str(self.block)


class MPS_GU2_P(MPS_P):
    """
    MPS with GU2 blocks.
    """
    def __init__(self, nl=1, nq=11, version=0, **kwargs):
        block = Block_GU2_P()
        super(MPS_GU2_P, self).__init__(block, nl, nq, irb=None, version=version, **kwargs)

    def __repr__(self):
        return "MPS_GU2_P({}, {}, {})".format(self.nl, self.nq, self.version)


class MPS_CZ_P(MPS_P):
    """
    MPS with CZ blocks.

    :param pm1: parameterization method for GU1. See gate.py
    :type pm1: int
    """
    def __init__(self, nl=1, nq=11, irb=None, version=0, pm1=0, **kwargs):
        self.pm1 = pm1
        block = Block_CZ_P(pm1=self.pm1)
        super(MPS_CZ_P, self).__init__(block, nl, nq, irb, version=version, **kwargs)

    def __repr__(self):
        return "MPS_CZ_P({}, {}, {}, {}, {})".format(self.nl, self.nq, repr(self.irb), self.version, self.pm1)


class MPS_CNOT_P(MPS_P):
    """
    MPS with CNOT blocks.

    :param pm1: parameterization method for GU1. See gate.py
    :type pm1: int
    """
    def __init__(self, nl=1, nq=11, irb=None, version=0, pm1=0, **kwargs):
        self.pm1 = pm1
        block = Block_CNOT_P(pm1=self.pm1)
        super(MPS_CNOT_P, self).__init__(block, nl, nq, irb, version=version, **kwargs)

    def __repr__(self):
        return "MPS_CNOT_P({}, {}, {}, {}, {})".format(self.nl, self.nq, repr(self.irb), self.version, self.pm1)



import setuptools


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='vqc_lite',
    version='0.1.0',
    author='Kevin Shen',
    author_email='kevinshen.abcd@gmail.com',
    description='vqc_lite: A reader-friendly Python implementation of Variational Quantum Circuits (VQC) based on Jax '
                'and Pennylane',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/kevinkayyy/VQC-Lite',
    license='Apache-2.0',
    packages=setuptools.find_packages(),
    install_requires=['numpy', 'matplotlib', 'pennylane', 'jax', 'jaxlib',  'optax', 'tqdm']
)



import numpy as np
import copy
from .gate import *
from .layer import Layer
from ..toolboxes.toolbox_statevector import distance, fidelity


class Circuit:
    """
    The circuit stack consists of 4 levels, which from bottom to top are: gate, block, layer and circuit.

    Circuits are at the top level of the stack.

    1. The most important attributes of a circuit are also "gs", "il" and "npl",
       which are formed by concatenating the ones of its layer components.
    2. The attribute "components" is a list of layers,
       which could be of different kinds.
    3. The attribute "filled_circuit" is a list of arrays,
       each corresponds to a fixed gate, or a parametrized gate filled with parameters.

    :param nq: the number of qubits in the circuit.
    :type nq: int
    :param pmd: the parameterization method dictionary, which defines how a gate should be parametrized if there are
                multiple ways to do so
    :type pmd: dictionary mapping str to int
    """
    def __init__(self, nq=4, **kwargs):
        self.nq = nq
        self.components = []  # layers
        self.gs, self.il, self.npl = [], [], []
        self.filled_circuit = None
        self.pmd = kwargs.get('pmd', {'GU1': None})

    def __repr__(self):
        return "Circuit({})".format(self.nq)

    def __str__(self):
        return "C-{}Q_".format(self.nq)

    def __add__(self, other):
        """
        We distinguish two cases for circuit addition.

        1. When we add with another circuit, a new circuit will be generated,
           the components of the two circuits will be added together, and the "pmd" of the circuit will be given by the
           other circuit.
        2. When we add the circuit with a layer, we will keep the circuit,
           the layer will simply be appended to the components. Then we will reassemble the circuit.

        :param other: circuit or layer to be added
        :type other: Circuit or Layer
        :return: new or modified circuit
        :rtype: Circuit
        """
        if isinstance(other, Circuit):
            new_circuit = copy.deepcopy(other)
            new_circuit.components = self.components + other.components
            new_circuit.assemble()
            return new_circuit
        if isinstance(other, Layer):
            self.components.append(other)
            self.assemble()
            return self

    def get_np(self):
        """
        Compute the total number of parameters in the block.

        :rtype: int
        """
        return sum(self.npl)

    def get_where_parametrized(self):
        """
        Return an array of integer indices, telling which gates of the gate sequence are parametrized gates

        :rtype: array of int
        """
        return np.nonzero(self.npl)[-1]

########################################################################################################################
    def assemble(self):
        """
        Automatically called during initialization to generate the three attributes.
        """
        self.npl = []
        self.gs = []
        self.il = []
        for component in self.components:
            self.npl += component.npl
            self.gs += component.gs
            self.il += component.il

########################################################################################################################
    def run_with_param_input(self, params):
        """
        Run the full circuit with parameter inputs for the parametrized gates.

        First compute the filled circuit and then perform a full forward contraction starting from state |0>.

        :param params: parameter inputs
        :type params: (jax) numpy array / list
        """
        psi = np.zeros(self.nq * [2])
        psi[tuple(self.nq * [0])] = 1

        self.fill_params(params)

        return self.forward_contraction(psi)

    def run_with_gate_input(self, gates):
        """
        Run the full circuit with gate inputs that would replace the parametrized gates and form the filled circuit
        together with the fixed gates in the gate sequence.

        First compute the filled circuit and then perform a full forward contraction starting from state |0>.

        :param gates: gate inputs
        :type gates: list of (jax) numpy arrays
        """
        psi = np.zeros(self.nq * [2])
        psi[tuple(self.nq * [0])] = 1

        self.fill_params(np.ones(self.get_np()))  # first create a trivial filled_circuit first
        for i in range(len(gates)):
            ig = self.get_where_parametrized()[i]  # find where the parametrized gates are
            self.filled_circuit[ig] = gates[i]  # replace the parametrized gates

        return self.forward_contraction(psi)

    def fill_params(self, params):
        """
        Form the filled_circuit by filling parameters into the parametrized gates.

        :param params: parameter inputs
        :type params: (jax) numpy array / list
        """
        param_count = 0
        gate_count = 0
        filled_circuit = []

        while gate_count < len(self.gs):

            gate_func = self.gs[gate_count]
            shape = tuple([2] * 2 ** len(self.il[gate_count]))  # compute the shape of the gate
            nparam = self.npl[gate_count]

            param = params[param_count: param_count + nparam]
            param_count += nparam

            if nparam > 0:  # not empty, parametrized gate
                if gate_func in self.pmd.keys():  # gate that has different parametrization methods
                    gate = globals()[gate_func](param, parametrization=self.pmd[gate_func])
                else:  # gate that has only one parameterization method
                    gate = globals()[gate_func](param).reshape(*shape)
            else:  # empty, fixed gate
                gate = globals()[gate_func]().reshape(*shape)

            filled_circuit.append(gate)

            gate_count += 1

        self.filled_circuit = filled_circuit

    def forward_contraction(self, psi_in, ng=np.inf):
        """
        Contract the filled circuit in forward direction until some stopping point, and compute the output state

        :param psi_in: A ket state. Usually |0>.
        :type psi_in: (jax) numpy array
        :param ng: The number of gates to contract before stop.
        :type ng: int

        :return: ket state after contraction
        :rtype: (jax) numpy array
        """
        leg_list = [[1], [2, 3]]  # for storing the legs of the gate during contraction

        def contraction(circuit, il, nq):
            psi = psi_in

            for gate_count in range(min(ng, len(circuit))):
                gate = circuit[gate_count]
                global_index = il[gate_count]

                n = len(global_index)

                psi = jnp.tensordot(gate, psi, [leg_list[n-1], global_index])

                order = list(range(n, nq))
                for i in range(n):
                    order.insert(global_index[i] % nq, i)

                psi = jnp.transpose(psi, axes=tuple(order))

            return psi

        return contraction(self.filled_circuit, self.il, self.nq)

    def backward_contraction(self, psi_out, ng=np.inf):
        """
        Contract the filled circuit in backward direction until some stopping point and compute the input state.

        :param psi_out: A bra state. When using in optimization, if psi is the target state, psi_out = psi.conjugate()
        :type psi_out: (jax) numpy array
        :param ng: The number of gates to contract before stop.
        :type ng: int
        :return: ket state after contraction
        :rtype: (jax) numpy array
        """
        leg_list = [[0], [0, 1]]  # for storing the legs of the gate during contraction

        def contraction(circuit, il, nq):
            psi = psi_out

            for gate_count in range(min(ng, len(circuit))):
                gate = circuit[len(circuit) - gate_count - 1]
                global_index = il[len(circuit) - gate_count - 1]

                n = len(global_index)

                psi = jnp.tensordot(gate, psi, [leg_list[n - 1], global_index])

                order = list(range(n, nq))
                for i in range(n):
                    order.insert(global_index[i] % nq, i)

                psi = jnp.transpose(psi, axes=tuple(order))

            return psi

        return contraction(self.filled_circuit, self.il, self.nq)

    def environment_contraction(self, psi_in, psi_out, ig):
        """
        Compute the environment tensor of a gate, by performing forward contraction from the input state, and backward
        contraction from the output state, toward the gate. This function is usually called during sweeping
        optimization.

        :param psi_in: A ket state. Usually |0>.
        :type psi_in: (jax) numpy array
        :param psi_out: A bra state. When using in optimization, if psi is the target state, psi_out = psi.conjugate()
        :type psi_out: (jax) numpy array
        :param ig: The index of the target gate. We start counting from 1!
        :type ig: int
        :return: The environment tensor
        :rtype: (jax) numpy array
        """
        psi_forward = self.forward_contraction(psi_in, ng=ig)  # perform forward contraction for ig gates
        psi_backward = self.backward_contraction(psi_out, ng=len(self.filled_circuit) - ig - 1)
        # perform backward contraction for len(circuit) - ig - 1 gates
        global_index = self.il[ig]  # The qubits the gate acts on.
        contract_index = [i for i in list(range(self.nq)) if i not in global_index]  # The indices to be contracted,
        # basically all the qubits that the gate does not act on.
        environment_tensor = jnp.tensordot(psi_forward, psi_backward, [contract_index, contract_index])

        return environment_tensor

########################################################################################################################
    def get_loss(self, params, psi, lf='td'):
        """
        Compute the "loss" of the output state of the VQC with parameter inputs compared to a target state psi.

        Often called during optimization for state preparation. Could also be used for more general purposes.

        :param params: parameter inputs
        :type params: (jax) numpy array / list
        :param psi: target state
        :type psi: (jax) numpy array
        :param lf: choice of loss function, currently supports:

                      1. coordinate-wise distance
                      2. squared coordinate-wise distance
                      3. 1 - fidelity
                      4. trace distance
        :type lf: int

        """
        if lf == 'd':  # Vector distance
            return self.get_distance_with(params, psi)
        if lf == 'd2':  # Vector distance squared
            return self.get_distance_with(params, psi)**2
        if lf == "1-f":  # 1 - fidelity
            return 1 - self.get_fidelity_with(params, psi)
        if lf == "td":  # Trace distance, related to fidelity for pure states
            return 2 * jnp.sqrt(1 - self.get_fidelity_with(params, psi))

    def get_distance_with(self, params, psi):
        """
        Compute the coordinate-wise distance between the output state of the VQC with parameter inputs and a target
        state psi.

        :param params: parameter inputs
        :type params: (jax) numpy array / list
        :param psi: target state
        :type psi: (jax) numpy array
        """
        psi_out = self.run_with_param_input(params)
        return distance(psi_out, psi)

    def get_fidelity_with(self, params, psi):
        """
        Compute the fidelity between the output state of the VQC with parameter inputs and a target state psi.

        :param params: parameter inputs
        :type params: (jax) numpy array / list
        :param psi: target state
        :type psi: (jax) numpy array
        """
        psi_out = self.run_with_param_input(params)
        return fidelity(psi_out, psi)

########################################################################################################################



import jax.numpy as jnp


def fidelity(psi1, psi2):
    """
    Compute the fidelity between 2 statevectors.

    :param psi1: the first statevector
    :type psi1: (jax) numpy array
    :param psi2: the second statevector
    :type psi2: (jax) numpy array
    :return: fidelity
    :rtype: float
    """
    return jnp.abs(jnp.sum(psi1 * psi2.conjugate())) ** 2


def distance(psi1, psi2):
    """
    Compute the coordinate-wise distance between 2 statevectors.

    :param psi1: the first statevector
    :type psi1: (jax) numpy array
    :param psi2: the second statevector
    :type psi2: (jax) numpy array
    :return: coordinate-wise distance
    :rtype: float
    """
    return jnp.abs(jnp.sqrt(jnp.sum(jnp.abs(psi1 - psi2) ** 2)))



import pennylane as qml
from ..circuit_stack.gate import GU1, GU2


class Block_P:
    """
    The Pennylane version of the circuit stack consists of 3 levels, which from bottom to top are: gate, block and
    circuit.

    Note that because of functions such as "qml.MPS" and "qml.broadcast", the arrangement of blocks could be
    easily handled at the circuit level for many common Ansätze. Thus, at the moment the "layer" level which exists in
    the other circuit stack is omitted here. However, when more complicated Ansätze come in the future, maybe it's
    better to also introduce the layer here.

    In the other stack where attributes are passed from one level to the next and only assembled and mapped to gate
    functions at the circuit level. In the Pennylane stack, in contrary, we directly pass callable gate functions called
    "circuit".

    A Block is the repeating unit in a layer. A block consists of >= 1 parmaterized or fixed gate(s).

    :param pm1: parametrization method for GU1
    :type pm1: int
    """
    def __init__(self, pm1=0):
        self.circuit = None
        self.np = 0
        self.assemble()
        self.pm1 = pm1

    def __repr__(self):
        return "Block_P({})".format(self.pm1)

    def __str__(self):
        return "B_"

    def assemble(self):
        """
        Automatically called during initialization to generate the circuit function.
        """
        pass


class Block_GU1_P(Block_P):
    """
    This block consists of only 1 general single qubit unitary. It's takes 3 parameters.
    """
    def __init__(self, pm1=0):
        super(Block_GU1_P, self).__init__(pm1=pm1)
        self.np = 3

    def __repr__(self):
        return "Block_GU1_P({})".format(self.pm1)

    def __str__(self):
        return "B-G1_"

    def assemble(self):
        def circuit(weights, wires):
            qml.QubitUnitary(GU1(weights, parametrization=self.pm1), wires=wires)
        self.circuit = circuit


########################################################################################################################

class Block_GU2_P(Block_P):
    """
    This block consists of only 1 general two qubit unitary (GU2). It takes 15 parameters.
    """
    def __init__(self):
        super(Block_GU2_P, self).__init__()
        self.np = 15
        self.pm1 = 'N/A'

    def __repr__(self):
        return "Block_G2QU_P()"

    def __str__(self):
        return "B-G2_"

    def assemble(self):
        def circuit(weights, wires):
            qml.QubitUnitary(GU2(weights), wires=wires)
        self.circuit = circuit


class Block_CZ_P(Block_P):
    """
    This block consists of a Controlled-Z gate, followed by 2 general single qubit unitaries, one on each qubit. It
    takes 6 parameters.
    """
    def __init__(self, pm1=0):
        super(Block_CZ_P, self).__init__(pm1=pm1)
        self.np = 6

    def __repr__(self):
        return "Block_CZ_P({})".format(self.pm1)

    def __str__(self):
        return "B-CZ_"

    def assemble(self):
        def circuit(weights, wires):
            qml.CZ(wires=wires)
            qml.QubitUnitary(GU1(weights[0:3], parametrization=self.pm1), wires=wires[0])
            qml.QubitUnitary(GU1(weights[3:6], parametrization=self.pm1), wires=wires[1])
        self.circuit = circuit


class Block_CNOT_P(Block_P):
    """
    This block consists of a Controlled-X gate, followed by 2 general single qubit unitaries (GU1), one on each qubit.
    It takes 6 parameters.
    """
    def __init__(self, pm1=0):
        super(Block_CNOT_P, self).__init__(pm1=pm1)
        self.np = 6

    def __repr__(self):
        return "Block_CNOT_P({})".format(self.pm1)

    def __str__(self):
        return "B-CNOT_"

    def assemble(self):
        def circuit(weights, wires):
            qml.CNOT(wires=wires)
            qml.QubitUnitary(GU1(weights[0:3], parametrization=self.pm1), wires=wires[0])
            qml.QubitUnitary(GU1(weights[3:6], parametrization=self.pm1), wires=wires[1])
        self.circuit = circuit



import jax
import jax.numpy as jnp
from jax.scipy.linalg import expm
from itertools import product

#  Pauli Basis #########################################################################################################

def Id1():
    """
    The identity matrix for the 1Q space
    """
    return jnp.eye(2)


def PauliX():
    """
    Pauli-X matrix / 1Q X-gate
    """
    return jnp.array([[0, 1], [1, 0]])


def PauliY():
    """
    Pauli-Y matrix / 1Q Y-gate
    """
    return jnp.array([[0, -1.j], [1.j, 0]])


def PauliZ():
    """
    Pauli-Z matrix / 1Q Z-gate
    """
    return jnp.array([[1, 0], [0, -1]])


def PauliBasis1():
    """
    1-Q Pauli basis stored in array
    """
    return jnp.array([Id1(), PauliX(), PauliY(), PauliZ()])


def PauliBasis2():
    """
    2-Q Pauli basis stored in array
    """
    return jnp.array([jnp.kron(pair[0], pair[1]) for pair in list(product(PauliBasis1(), PauliBasis1()))])


#  Parametrized 1Q gates ###############################################################################################

def RX(x):
    """
    The parametrized RX-gate

    :param x: the rotation angle
    :type x: float
    """
    return jnp.array([[jnp.cos(x / 2), -1j * jnp.sin(x / 2)], [-1j * jnp.sin(x / 2), jnp.cos(x / 2)]])


def RY(x):
    """
    The parametrized RY-gate

    :param x: the rotation angle
    :type x: float
    """
    return jnp.array([[jnp.cos(x / 2), -jnp.sin(x / 2)], [jnp.sin(x / 2), jnp.cos(x / 2)]])


def RZ(x):
    """
    The parametrized RY-gate

    :param x: the rotation angle
    :type x: float
    """
    return jnp.array([[jnp.exp(-1j * x / 2), 0], [0, jnp.exp(1j * x / 2)]])


def U3_Rot(params):
    """
    General 1Q unitary. Same parametrization method as "U3Gate" from Qiskit.

    :param params: the 3 rotation angles, theta, phi, and lambda.
    :type params: (jax) numpy array / list of float
    """
    t, p, l = params
    gate = jnp.array([[jnp.cos(t / 2), -jnp.exp(1j * l) * jnp.sin(t / 2)],
                      [jnp.exp(1j * p) * jnp.sin(t / 2), jnp.exp(1j * (p + l)) * jnp.cos(t / 2)]])
    return gate


def U3_Pauli(params):
    """
    General 1Q unitary, parametrized by Pauli basis decomposition of the exponent Hermitian matrix.

    :param params: the 3 coefficients of Pauli basis elements (excluding that of Id, which would only contributes to an
                   additional global phase).
    :type params: (jax) numpy array / list of float
    """
    x, y, z = params
    n = jnp.linalg.norm(params)
    a = jnp.cos(n / 2) - 1j * z / n * jnp.sin(n / 2)
    b = -y / n * jnp.sin(n / 2) - 1j * x / n * jnp.sin(n / 2)
    c = y / n * jnp.sin(n / 2) - 1j * x/n * jnp.sin(n / 2)
    return jnp.array([[a, b], [c, jnp.conjugate(a)]])


def GU1(params, parametrization=0):
    """
    A wrapper function for the various functions realizing the general 1Q unitary. There are many ways to parametrize a
    general single qubit unitary with 3 parameters.

    :param params: An array of 3 elements.
    :type params: (jax) numpy array / list of float
    :param parametrization: An integer that defines the parametrization method. Currently, 5 methods are supported.

                            0. Pauli basis parametrization
                            1. Euler angle parametrization, RZ - RY - RZ. Same as "qml.Rot" from Pennylane
                            2. Euler angle parametrization, RY - RZ - RY.
                            3. Euler angle parametrization, RX - RZ - RX.
                            4. Euler angle parametrization, qiskit version
    :type parametrization: int
    """
    if parametrization == 0:  # recommended interval for sampling: [0, 4pi]
        return U3_Pauli(params)
    if parametrization == 1:  # recommended interval for sampling: [0, 4pi]
        return RZ(params[2]) @ RY(params[1]) @ RZ(params[0])
    if parametrization == 2:  # recommended interval for sampling: [0, 4pi]
        return RY(params[2]) @ RZ(params[1]) @ RY(params[0])
    if parametrization == 3:  # recommended interval for sampling: [0, 4pi]
        return RX(params[2]) @ RZ(params[1]) @ RX(params[0])
    if parametrization == 4:  # recommended interval for sampling: [0, 4pi]
        return U3_Rot(params)


#  Fixed 2QG ###########################################################################################################

def CNOT():
    """
    The circuit stack consists of 4 levels, which from bottom to top are: gate, block, layer and circuit.

    Gates are at the bottom level of the stack. Functions in this module realize gates that can be either parametrized
    or unparametrized. Parametrized gates take parameters as input, while unparametrized gates are without input.
    Functions of both kinds output a unitary matrix in jax numpy array form.

    Controlled X gates. Native entangling gate of many superconducting quantum computers.
    """
    return jnp.array([[1, 0, 0, 0],
                      [0, 1, 0, 0],
                      [0, 0, 0, 1],
                      [0, 0, 1, 0]])


def CZ():
    """
    Controlled Z gates. Native entangling gate of many superconducting quantum computers.
    """
    return jnp.array([[1, 0, 0, 0],
                      [0, 1, 0, 0],
                      [0, 0, 1, 0],
                      [0, 0, 0, -1]])


#  Parametrized 2QG ####################################################################################################

def GU2(params):
    """
    General 2Q unitary parametrized by 15 coefficients of the 2Q Pauli basis (excluding Id)

    :param params: An array of 15 elements.
    :type params: (jax) numpy array / list of float
    """
    return expm(-0.5j * jnp.tensordot(params, PauliBasis2()[1:], [0, 0]))


#  Others ##############################################################################################################

def Haar_Random(nq, key):
    """
    Randomly generates a general nq-qubit unitary, with respect to Haar measure, by using QR-decomposition.
    Ref: https://arxiv.org/abs/math-ph/0609050


    :param nq: the number of qubits
    :type nq: int
    :param key: the random key
    :type key: jax key
    """
    N = 2 ** nq
    key, subkey = jax.random.split(key)
    M = jax.random.normal(key, shape=(N, N)) + 1j * jax.random.normal(subkey, shape=(N, N))
    Q, R = jnp.linalg.qr(M)
    D = jnp.diagonal(R)
    L = D / jnp.absolute(D)
    U = jnp.multiply(R, L)

    return U
